#!/usr/bin/env python

import os
from argparse import ArgumentParser

from dtags.help import HelpFormatter
from dtags.colors import PINK, CYAN, YELLOW, CLEAR
from dtags.chars import TAG_NAME_CHARS
from dtags.config import load_tags, save_tags
from dtags.utils import expand_path, shrink_path

cmd_description = """
dtags - tag directories

e.g. the command {y}tag ~/foo ~/bar @a @b ~/baz @a @c{x}:

    adds to directory {c}~/foo{x} tags {p}@a @b{x}
    adds to directory {c}~/bar{x} tags {p}@a @b{x}
    adds to directory {c}~/baz{x} tags {p}@a @c{x}

""".format(p=PINK, c=CYAN, y=YELLOW, x=CLEAR)

msg = 'Added tag {p}{{}}{x} to {c}{{}}{x}'.format(p=PINK, c=CYAN, x=CLEAR)


def main():
    tag_to_paths = load_tags()
    parser = ArgumentParser(
        prog='tag',
        usage='tag [[paths] [tags]...]',
        description=cmd_description,
        formatter_class=HelpFormatter
    )
    parser.add_argument(
        'arguments',
        type=str,
        nargs='+',
        metavar='[paths] [tags]',
        help='directory paths and tag names'
    )
    parsed = parser.parse_args()

    # Tracking variables and collectors
    updates = []
    arg_index = 0
    parsing_paths = True
    tags_collected = set()
    paths_collected = set()

    # Iterate through the arguments and pair up tags with paths
    while arg_index < len(parsed.arguments):
        arg = parsed.arguments[arg_index]
        if parsing_paths and arg.startswith('@'):
            if len(paths_collected) == 0:
                parser.error('expecting paths before {}'.format(arg))
            parsing_paths = False
        elif parsing_paths and not arg.startswith('@'):
            paths_collected.add(arg)
            arg_index += 1
        elif not parsing_paths and arg.startswith('@'):
            tag_name_has_alphabet = False
            for ch in arg[1:]:
                if ch not in TAG_NAME_CHARS:
                    parser.error('bad char {} in tag name {}'.format(ch, arg))
                tag_name_has_alphabet |= ch.isalpha()
            if not tag_name_has_alphabet:
                parser.error('no alphabets in tag name {}'.format(arg))
            tags_collected.add(arg)
            arg_index += 1
        else:
            updates.append((tags_collected, paths_collected))
            tags_collected, paths_collected = set(), set()
            parsing_paths = True
    if parsing_paths:
        parser.error('expecting a tag name')
    updates.append((tags_collected, paths_collected))

    # Apply updates and message
    messages = set()
    for tags, paths in updates:
        valid_paths = [p for p in paths if os.path.isdir(p)]
        if len(valid_paths) == 0:
            continue
        for tag in tags:
            if tag not in tag_to_paths:
                tag_to_paths[tag] = {}
            for path in valid_paths:
                full_path = expand_path(path)
                short_path = shrink_path(path)
                tag_to_paths[tag][full_path] = short_path
                messages.add(msg.format(tag, full_path))
    save_tags(tag_to_paths)
    if messages:
        print('\n'.join(messages))



"""Formatter classes used with the argparse library."""

from argparse import PARSER, RawDescriptionHelpFormatter


class HelpFormatter(RawDescriptionHelpFormatter):
    """This makes help menus look nicer."""

    def _format_action(self, action):
        parts = super(RawDescriptionHelpFormatter, self)._format_action(action)
        if action.nargs == PARSER:
            return '\n'.join(parts.split('\n')[1:])
        return parts



#!/usr/bin/env python

import os
import argparse

from dtags.help import HelpFormatter
from dtags.colors import PINK, CYAN, YELLOW, CLEAR
from dtags.config import load_tags, save_tags
from dtags.utils import expand_path

cmd_description = """
dtags - untag directories

e.g. the command {y}untag ~/foo ~/bar @a @b ~/baz @c @d{x}:

    removes from directory {c}~/foo{x} tags {p}@a @b{x}
    removes from directory {c}~/bar{x} tags {p}@a @b{x}
    removes from directory {c}~/baz{x} tags {p}@c @d{x}

""".format(p=PINK, c=CYAN, y=YELLOW, x=CLEAR)

msg = 'Removed tag {p}{{}}{x} from {c}{{}}{x}'.format(p=PINK, c=CYAN, x=CLEAR)


def main():
    tag_to_paths = load_tags()
    parser = argparse.ArgumentParser(
        prog="untag",
        usage="untag [options] [[paths] [tags]...]",
        description=cmd_description,
        formatter_class=HelpFormatter
    )
    parser.add_argument(
        '-a', '--all',
        help='untag all tags or paths',
        action='store_true'
    )
    parser.add_argument(
        'arguments',
        type=str,
        nargs=argparse.REMAINDER,
        metavar='[paths] [tags]',
        help='directory paths and tag names'
    )
    parsed = parser.parse_args()

    if parsed.all:
        # Collect all the paths and tags to remove
        messages = set()
        tags_to_remove = set()
        paths_to_remove = set()
        for arg in parsed.arguments:
            if arg in tag_to_paths:
                tags_to_remove.add(arg)
            elif os.path.isdir(arg):
                paths_to_remove.add(expand_path(arg))

        # Untag all the paths specified
        for tag, paths in tag_to_paths.items():
            for path in [p for p in paths if p in paths_to_remove]:
                paths.pop(path)
                messages.add(msg.format(tag, path))
            if len(paths) == 0:
                tags_to_remove.add(tag)

        # Remove all the tags specified
        for tag in tags_to_remove:
            paths = tag_to_paths.pop(tag)
            for path in paths:
                messages.add(msg.format(tag, expand_path(path)))

        # Save the updated tags and print messages
        save_tags(tag_to_paths)
        if messages:
            print('\n'.join(messages))
    else:
        # Initialize tracking variables and collectors
        updates = []
        arg_index = 0
        parsing_paths = True
        tags_collected = set()
        paths_collected = set()

        # Iterate through the arguments and pair up tags with paths
        while arg_index < len(parsed.arguments):
            arg = parsed.arguments[arg_index]
            if parsing_paths and arg.startswith('@'):
                if len(paths_collected) == 0:
                    parser.error('excepting paths before {}'.format(arg))
                parsing_paths = False
            elif parsing_paths and not arg.startswith('@'):
                paths_collected.add(arg)
                arg_index += 1
            elif not parsing_paths and arg.startswith('@'):
                tags_collected.add(arg)
                arg_index += 1
            else:
                updates.append((tags_collected, paths_collected))
                tags_collected, paths_collected = set(), set()
                parsing_paths = True
        if parsing_paths:
            parser.error('expecting a tag name')
        updates.append((tags_collected, paths_collected))

        # Apply updates and collect messages to print
        messages = set()
        for tags, paths in updates:
            for tag in tags:
                if tag not in tag_to_paths:
                    continue
                for path in paths:
                    if not os.path.isdir(path):
                        continue
                    full_path = expand_path(path)
                    if full_path in tag_to_paths[tag]:
                        tag_to_paths[tag].pop(full_path)
                        messages.add(msg.format(tag, full_path))
                if len(tag_to_paths[tag]) == 0:
                    # Remove the tag completely if it has no paths
                    tag_to_paths.pop(tag)

        # Save the updated tags and print messages
        save_tags(tag_to_paths)
        if messages:
            print('\n'.join(messages))



#!/usr/bin/env python

import os
import sys
import errno
import signal
import atexit
import argparse
import tempfile
import subprocess
from collections import OrderedDict

from dtags.colors import *
from dtags.completion import ChoicesCompleter, autocomplete
from dtags.config import load_tags
from dtags.help import HelpFormatter
from dtags.utils import expand_path

tmp_file = None
process = None
processes = []
cmd_description = """
dtags - run commands in directories

e.g. the command {y}run @a @b ~/foo ~/bar ls -la{x}:

    runs {y}ls -la{x} in all directories with tag {p}@a{x}
    runs {y}ls -la{x} in all directories with tag {p}@b{x}
    runs {y}ls -la{x} in directory {c}~/foo{x}
    runs {y}ls -la{x} in directory {c}~/bar{x}

""".format(p=PINK, c=CYAN, y=YELLOW, x=CLEAR)


def _print_header(tag, path):
    """Print the run information header."""
    tail = ' ({}{}{}):'.format(PINK, tag, CLEAR) if tag else ':'
    print('{}{}{}{}'.format(CYAN, path, tail, CLEAR))


def _print_error(cmd, err):
    """Print the error message."""
    print('Failed to run {}{}{}: {}'.format(YELLOW, cmd, CLEAR, err))


def _print_exit_code(exit_code):
    """Print the exit code from process."""
    print('{}Exit Code: {}{}'.format(RED, exit_code, CLEAR))


def _send_sigterm(target_process):
    """Send sigterm to the target child process."""
    try:
        os.killpg(os.getpgid(target_process.pid), signal.SIGKILL)
    except OSError as err:
        # If the process is already killed for some reason, do nothing
        if err.errno != errno.ESRCH:
            raise


def _cleanup_resources():
    """Kill child processes and remove temporary files."""
    global tmp_file, process, processes
    if tmp_file is not None:
        tmp_file.close()
    if process is not None:
        _send_sigterm(process)
    for _, _, proc, tmp in processes:
        _send_sigterm(proc)
        tmp.close()


def _kill_signal_handler(signum, frame):
    """Clean up resources when sigterm or sigint is received."""
    _cleanup_resources()
    sys.exit(127 + signum)


def main():
    global tmp_file, process, processes

    # Ensure that child processes are cleaned up
    signal.signal(signal.SIGINT, _kill_signal_handler)
    signal.signal(signal.SIGTERM, _kill_signal_handler)
    atexit.register(_cleanup_resources)

    tag_to_paths = load_tags()
    parser = argparse.ArgumentParser(
        prog='run',
        description=cmd_description,
        usage='run [options] [targets] command',
        formatter_class=HelpFormatter
    )
    parser.add_argument(
        '-p', '--parallel',
        help='run the commands in parallel',
        action='store_true'
    )
    parser.add_argument(
        '-e', '--exit-codes',
        help='display exit codes',
        action='store_true'
    )
    parser.add_argument(
        'arguments',
        type=str,
        nargs=argparse.REMAINDER,
        metavar='[targets]',
        help='tags or paths to run the command for'
    ).completer = ChoicesCompleter(tag_to_paths.keys())
    autocomplete(parser)
    parsed = parser.parse_args()

    # Separate the targets from the command
    index = 0
    targets = OrderedDict()
    while index < len(parsed.arguments):
        target = parsed.arguments[index]
        if target in tag_to_paths:
            for path in sorted(tag_to_paths[target].keys()):
                targets[path] = target
            index += 1
        elif target.startswith('@'):
            parser.error('unknown tag {}'.format(target))
        elif os.path.isdir(target):
            path = expand_path(target)
            if path not in targets:
                targets[path] = None
            index += 1
        else:
            break  # beginning of the command
    command = ' '.join(parsed.arguments[index:])
    if not (targets and command):
        parser.error('too few arguments')

    exit_code = 0
    if parsed.parallel:
        # Run the command in parallel
        for path, tag in targets.items():
            tmp_file = tempfile.TemporaryFile(mode='w+t')
            process = subprocess.Popen(
                command,
                cwd=path,
                stdout=tmp_file,
                stderr=tmp_file,
                shell=True,
                preexec_fn=os.setsid
            )
            processes.append((path, tag, process, tmp_file))
        for path, tag, process, tmp_file in processes:
            child_exit_code = process.wait()
            tmp_file.seek(0)
            _print_header(tag, path)
            print(tmp_file.read())
            tmp_file.close()
            if child_exit_code != 0:
                exit_code = 1
            if parsed.exit_codes:
                _print_exit_code(child_exit_code)
            print('')
    else:
        # Run the command sequentially
        for path, tag in targets.items():
            _print_header(tag, path)
            child_exit_code = subprocess.call(
                command,
                cwd=path,
                shell=True,
                stderr=sys.stdout.fileno()
            )
            if child_exit_code != 0:
                exit_code = 1
            if parsed.exit_codes:
                _print_exit_code(child_exit_code)
            print('')
    sys.exit(exit_code)



#!/usr/bin/env python

import os
import sys
import json
import shutil
import subprocess

from collections import defaultdict
from argparse import ArgumentParser

from dtags.config import TAGS_FILE_PATH
from dtags.colors import PINK, CYAN, YELLOW, CLEAR
from dtags.completion import ChoicesCompleter, autocomplete
from dtags.config import load_tags, check_tags, save_tags
from dtags.help import HelpFormatter
from dtags.utils import halt, expand_path, shrink_path, msgify, safe_remove

cmd_description = """
dtags - display tags and tagged directories

e.g. the command {y}tags @a @b ~/foo ~/bar @c{x}:

    displays tags {p}@a @b @c{x}
    displays all tags with directory {c}~/foo{x}
    displays all tags with directory {c}~/bar{x}

""".format(p=PINK, c=CYAN, y=YELLOW, x=CLEAR)


def main():
    tag_to_paths = load_tags()
    path_to_tags = defaultdict(set)
    for tag, paths in tag_to_paths.items():
        for path in paths.keys():
            path_to_tags[path].add(tag)

    parser = ArgumentParser(
        prog='tags',
        description=cmd_description,
        usage='tags [options] [paths] [tags]',
        formatter_class=HelpFormatter
    )
    parser.add_argument(
        '-e', '--edit',
        action='store_true',
        help='Edit the tags directly using an editor'
    )
    parser.add_argument(
        '-x', '--expand',
        action='store_true',
        help='Expand the directory paths'
    )
    parser.add_argument(
        '-r', '--reverse',
        help='display the reverse mapping',
        action='store_true'
    )
    parser.add_argument(
        '-j', '--json',
        help='display the raw JSON',
        action='store_true'
    )
    parser.add_argument(
        'search_terms',
        type=str,
        nargs='*',
        metavar='[paths] [tags]',
        help='tag and directory paths to filter by'
    ).completer = ChoicesCompleter(tag_to_paths.keys())
    autocomplete(parser)
    parsed = parser.parse_args()

    if parsed.edit:
        if parsed.search_terms:
            raise parser.error('no arguments allowed with option -e/--edit')
        edit_file_path = TAGS_FILE_PATH + '.edit'
        shutil.copy2(TAGS_FILE_PATH, edit_file_path)
        subprocess.call([os.environ.get('EDITOR', 'vi'), edit_file_path])
        new_tag_to_paths = save_error = None
        with open(edit_file_path, 'r') as edit_file:
            try:
                new_tag_to_paths = json.load(edit_file)
            except ValueError as err:
                save_error = 'Failed to save tags: {}'.format(msgify(err))
            else:
                try:
                    check_tags(new_tag_to_paths)
                except ValueError as err:
                    save_error = 'Failed to save tags: {}'.format(err)
        safe_remove(edit_file_path)
        if save_error is not None:
            halt(save_error)
        save_tags({
            tag: {expand_path(p): shrink_path(p) for p in paths}
            for tag, paths in new_tag_to_paths.items()
        })
        print('New tags saved successfully')
        sys.exit(0)

    if len(tag_to_paths) == 0:
        print('No tags defined')
        sys.exit(0)

    # Filter by any given tags and paths
    # TODO optimize here if possible
    if not parsed.search_terms:
        if parsed.expand:
            filtered = {t: ps.keys() for t, ps in tag_to_paths.items()}
        else:
            filtered = {t: ps.values() for t, ps in tag_to_paths.items()}
    else:
        filtered = {}
        for term in parsed.search_terms:
            if term in tag_to_paths:
                if parsed.expand:
                    filtered[term] = tag_to_paths[term].keys()
                else:
                    filtered[term] = tag_to_paths[term].values()
            elif os.path.isdir(term):
                term = expand_path(term)
                if term in path_to_tags:
                    for tag in path_to_tags[term]:
                        if parsed.expand:
                            filtered[tag] = tag_to_paths[tag].keys()
                        else:
                            filtered[tag] = tag_to_paths[tag].values()

    if parsed.json:
        formatted = {tag: sorted(paths) for tag, paths in filtered.items()}
        print(json.dumps(formatted, sort_keys=True, indent=4))
    elif parsed.reverse:
        reverse = defaultdict(set)
        for tag, paths in filtered.items():
            for path in paths:
                reverse[path].add(tag)
        for path, tags in reverse.items():
            print('{}{}{}'.format(CYAN, path, CLEAR))
            print('{}{}{}\n'.format(PINK, ' '.join(sorted(tags)), CLEAR))
    else:
        for tag, paths in sorted(filtered.items()):
            print('{}{}{}'.format(PINK, tag, CLEAR))
            print('{}{}{}\n'.format(CYAN, '\n'.join(sorted(paths)), CLEAR))



"""All the things related to auto-completion."""

from argcomplete import CompletionFinder


class DTagsCompletionFinder(CompletionFinder):
    """Completion finder for DTags."""

    def _get_option_completions(self, parser, cword_prefix):
        """Skip auto-completion for optional arguments."""
        return []


class ChoicesCompleter(object):
    """Completer initialized with a set of all possible choices."""

    def __init__(self, choices=None):
        self.choices = map(str, choices) if choices else []

    def __call__(self, prefix, **kwargs):
        return (c for c in self.choices if c.startswith(prefix))

autocomplete = DTagsCompletionFinder()



from setuptools import setup, find_packages

setup(
    name='dtags',
    version='1.0.6',
    description='Directory tags for lazy programmers',
    author='Joohwan Oh',
    author_email='joohwan.oh@outlook.com',
    url='https://github.com/joowani/dtags',
    download_url='https://github.com/joowani/dtags/tarball/1.0.6',
    packages=find_packages(),
    install_requires=['argcomplete'],
    entry_points={
        'console_scripts': [
            'run = dtags.commands.run:main',
            'tag = dtags.commands.tag:main',
            'tags = dtags.commands.tags:main',
            'untag = dtags.commands.untag:main',
        ],
    },
)



import os
import json

from dtags.chars import TAG_NAME_CHARS
from dtags.utils import halt, expand_path, is_dict, is_list, is_str

TAGS_FILE_PATH = expand_path('~/.dtags')
TEMP_FILE_PATH = TAGS_FILE_PATH + '.tmp'


def load_tags():
    """Load the tags from disk.

    Return a dictionary which maps tag names to sub-dictionaries, which in
    turn maps fully expanded paths to unexpanded paths like this:

    {
        "@app" : {
            "/home/user/app/backend": "~/app/backend",
            "/home/user/app/frontend": "~/app/frontend",
        },
        "@frontend": {
            "/home/user/app/frontend": "~/app/frontend",
        },
        "@backend": {
            "/home/user/app/backend": "~/app/backend",
        }
        ...
    }

    :return: the tags loaded from the disk
    """
    if not os.path.exists(TAGS_FILE_PATH):
        try:
            with open(TAGS_FILE_PATH, 'w') as open_file:
                json.dump({}, open_file)
            return {}
        except (IOError, OSError):
            halt('Failed to initialize {}'.format(TAGS_FILE_PATH))
    else:
        try:
            with open(TAGS_FILE_PATH, 'r') as open_file:
                json_str = open_file.read().strip()
                if not json_str:
                    return {}
                tag_data = json.loads(json_str)
                if not tag_data:
                    return {}
                else:
                    return {
                        tag: {expand_path(path): path for path in paths}
                        for tag, paths in tag_data.items()
                    }
        except (ValueError, IOError, OSError):
            # TODO better error handling and messaging here
            halt('Failed to load {}'.format(TAGS_FILE_PATH))


def save_tags(tags):
    """Save the tags to disk.

    Convert the incoming tags dictionary to JSON and dump the content to the
    file. For portability, only the unexpanded paths are saved.

    :param tags: tags to save to disk
    """
    try:
        with open(TEMP_FILE_PATH, 'w') as open_file:
            json.dump(
                {tag: sorted(paths.values()) for tag, paths in tags.items()},
                open_file,
                indent=4,
                sort_keys=True
            )
    except IOError:
        # TODO better error handling and messaging here
        halt('Failed to write to {}'.format(TAGS_FILE_PATH))
    else:
        # Overwrite the old file with the new
        os.rename(TEMP_FILE_PATH, TAGS_FILE_PATH)


def check_tags(tags):
    """Check the tags for correctness.

    :param tags: tags to check for correctness
    :raise: ValueError
    """
    if not is_dict(tags):
        raise ValueError('expecting a JSON object')
    for tag, paths in tags.items():
        if not (is_str(tag) and tag.startswith('@')):
            raise ValueError('invalid tag name {}'.format(tag))
        tag_name_has_alphabet = False
        for ch in tag[1:]:
            if ch not in TAG_NAME_CHARS:
                raise ValueError('bad char {} in tag name {}'.format(ch, tag))
            tag_name_has_alphabet |= ch.isalpha()
        if not tag_name_has_alphabet:
            raise ValueError('no alphabets in tag name {}'.format(tag))
        if not (is_list(paths) and len(paths) > 0):
            raise ValueError('expecting a non-empty list for {}'.format(tag))
        for path in paths:
            if not (is_str(path) and os.path.isdir(expand_path(path))):
                raise ValueError(
                    'invalid directory path {} for {}'.format(path, tag)
                )



from __future__ import print_function

import os
import sys
import errno
import importlib
from sys import stderr

from collections import Mapping
# Importing this way for python 2 and 3 compatibility
try:
    urllib = importlib.import_module('urllib.parse')
except ImportError:
    urllib = importlib.import_module('urllib')
try:
    builtins = importlib.import_module('__builtin__')
except ImportError:
    builtins = importlib.import_module('builtins')


def is_str(obj):
    """Check if the object is a str or a unicode

    :param obj: object to check
    :return: True if str or unicode, else False
    """
    base_str = getattr(builtins, 'basestring', None)
    return isinstance(obj, base_str) if base_str else isinstance(obj, str)


def is_list(obj):
    """Check if the object is an iterable

    :param obj: object to check
    :return: True if iterable, else False
    """
    return type(obj) == list


def is_dict(obj):
    """Check if the object is a mapping

    :param obj: object to check
    :return: True if mapping, else False
    """
    return isinstance(obj, Mapping)


def halt(message, exit_code=errno.EPERM):
    """Print the message to stderr and exit with the exit code.

    :param message: error message
    :param exit_code: linux exit code
    """
    print(message, file=stderr)
    sys.exit(exit_code)


def expand_path(path):
    """Fully expand a directory path.

    :param path: a valid directory path
    :return: expanded path
    """
    return os.path.realpath(os.path.expanduser(path))


def shrink_path(path):
    """Shrink the 'home' portion of the path to '~'.

    :param path: a valid directory path
    :return: shortened path
    """
    path = expand_path(path)
    home = expand_path('~')
    return path.replace(home, '~') if path.startswith(home) else path


def msgify(exception):
    """Format the exception message.

    1. Convert any capital characters to lowercase
    2. Strip any periods or newlines

    :param exception: the exception whose message to msgify
    :return: msgified string
    """
    string = str(exception)
    string = string if len(string) == 0 else string[0].lower() + string[1:]
    return string.strip('.')


def safe_remove(path):
    """Remove a file safely.

    If the file does not exist, then do nothing.

    :param path: path of the file to remove
    """
    try:
        os.remove(path)
    except OSError:
        pass
